import requests
import json
from rfc3986 import urlparse
import socket

from .logging import logger

infostring = "Please contact an administrator for more information."
# From the docs:
# > It’s a good practice to set connect timeouts to slightly larger than a multiple of 3,
# > which is the default TCP packet retransmission window.
TIMEOUT = 3.05


def deploy(mc_endpoint, token, verify=True):
    endpoint = mc_endpoint + "/user/deploy"
    headers = {"Authorization": f"Bearer {token}"}

    resp = requests.get(endpoint, headers=headers, verify=verify)

    from_cache = getattr(resp, "from_cache", False)
    if from_cache:
        logger.debug(f"Using cached response for {endpoint}")

    return resp


def get_status(mc_endpoint, token, verify=True):
    endpoint = mc_endpoint + "/user/get_status"
    headers = {"Authorization": f"Bearer {token}"}

    resp = requests.get(endpoint, headers=headers, verify=verify)

    from_cache = getattr(resp, "from_cache", False)
    if from_cache:
        logger.debug(f"Using cached response for {endpoint}")

    return resp


def generate_otp(mc_endpoint, token, verify=True):
    endpoint = mc_endpoint + "/user/generate_otp"
    headers = {"Authorization": f"Bearer {token}"}

    resp = requests.get(endpoint, headers=headers, verify=verify)

    from_cache = getattr(resp, "from_cache", False)
    if from_cache:
        logger.debug(f"Using cached response for {endpoint}")

    return resp


def info(mc_endpoint, verify=True):
    endpoint = mc_endpoint + "/info"

    resp = requests.get(endpoint, verify=verify)

    from_cache = getattr(resp, "from_cache", False)
    if from_cache:
        logger.debug(f"Using cached response for {endpoint}")

    return resp


def info_authorisation(mc_endpoint, token, verify=True):
    endpoint = mc_endpoint + "/info/authorisation"
    headers = {"Authorization": f"Bearer {token}"}

    resp = requests.get(endpoint, headers=headers, verify=verify)

    from_cache = getattr(resp, "from_cache", False)
    if from_cache:
        logger.debug(f"Using cached response for {endpoint}")

    return resp


def get_info(mc_endpoint, verify=True):
    try:
        resp = info(mc_endpoint, verify=verify)
        if resp.status_code == requests.codes.ok:  # pylint: disable=no-member
            return resp.json()
        else:
            resp.raise_for_status()
    except Exception as e:
        logger.debug(f"Something went wrong: {e}")
        logger.error("Failed to get service info")
    return None


def get_supported_ops(mc_endpoint, verify=True):
    service_info = get_info(mc_endpoint, verify)
    if service_info is not None:
        supported_ops = service_info.get("supported_OPs", [])
        # backward compat with old motley_cue
        if supported_ops == []:
            supported_ops = service_info.get("supported OPs", [])
        return supported_ops
    return []


def get_authorisation_info(mc_endpoint, token, verify=True):
    try:
        resp = info_authorisation(mc_endpoint, token, verify=verify)
        return resp.json()
    except Exception as e:
        logger.debug(f"Something went wrong: {e}")
        logger.error("Failed to get authorisation info from service")
    return None


def get_audience(mc_endpoint, token, verify=True):
    try:
        resp = info_authorisation(mc_endpoint, token, verify=verify)
        return resp.json().get("audience", None)
    except Exception as e:
        logger.debug("Failed to get audience from service, assuming no specific audience needed.")
    return None


def get_local_status(mc_endpoint, token, verify=True):
    try:
        resp = get_status(mc_endpoint, token, verify=verify)
        if resp.status_code == requests.codes.ok:  # pylint: disable=no-member
            output = resp.json()
            state = output["state"]
            status_string = None
            if state == "suspended":
                status_string = f"Your account on service is suspended, you might not be able to login. {infostring}"
                status_string += f'\nLocal username: {output["message"].split()[1]}'
            elif state == "limited":
                status_string = f"Your account on service has limited capabilities, but you might still be able to login. {infostring}"
                status_string += f'\nLocal username: {output["message"].split()[1]}'
            elif state == "pending":
                status_string = (
                    f"Your account creation on service is still pending approval. {infostring}"
                )
            elif state == "unknown":
                status_string = f"Your account on service is in an undefined state. {infostring}"
            elif state == "not_deployed":
                status_string = f"Your account on service is not deployed, but it will be created on the first login if authorised."
            elif state == "deployed":
                status_string = f"Your account on service is deployed."
                status_string += f'\nLocal username: {output["message"].split()[1]}'
            if status_string:
                return status_string
            else:
                # should not happen
                return "Failed to get more information about your local account."
        else:
            return resp.text
    except Exception as e:
        logger.debug(f"Something went wrong: {e}")
        logger.error("Failed to get local account info from service")
    return None


def _get_username_from_message(message):
    try:
        return message.split()[1]
    except Exception as e:
        logger.debug(f"Something went wrong: {e}")
        logger.error("Failed to get SSH username from server")

def local_username(mc_endpoint, token, verify=True):
    resp = get_status(mc_endpoint, token, verify=verify)
    if resp.status_code == requests.codes.ok:  # pylint: disable=no-member
        output = resp.json()
        state = output["state"]
        logger.info(f"State of your local account: {state}")
        if state == "suspended":
            logger.warning(
                f"Your account on service is suspended, you might not be able to login. {infostring}"
            )
            return _get_username_from_message(output["message"])
        elif state == "limited":
            logger.warning(
                f"Your account on service has limited capabilities, but you might still be able to login. {infostring}"
            )
            return _get_username_from_message(output["message"])
        elif state in ["unknown", "not_deployed", "deployed", "pending"]:
            if state == "unknown":
                logger.warning(
                    "Your account on service is in an undefined state. Will try redeploying..."
                )
            elif state == "not_deployed":
                logger.info("Creating local account...")
            elif state == "deployed":
                logger.info("Updating local account...")
            elif state == "pending":
                logger.info("Your account creation on service is still pending approval. Sending another request for deployment...")
            resp = deploy(mc_endpoint, token, verify=verify)
            if resp.status_code == requests.codes.ok:  # pylint: disable=no-member
                new_state = resp.json()["state"]
                if new_state == "pending":
                    message = resp.json().get("message", "Your account creation on service is pending approval.")
                    raise Exception(
                        f"{message} {infostring}"
                    )
                elif new_state == "deployed":
                    logger.debug(json.dumps(resp.json(), indent=2))
                    return resp.json()["credentials"]["ssh_user"]
                return _get_username_from_message(resp.json()["message"])
            elif state == "deployed":
                logger.warning(
                    f"Failed on redeploy. Some of your user information might be outdated."
                )
                return _get_username_from_message(output["message"])
            else:
                resp_dict = json.loads(resp.text)
                try:
                    logger.error(
                        f'Failed on deploy: [HTTP {resp.status_code}] [state={resp_dict["state"]}] {resp_dict["message"]}'
                    )
                except Exception:
                    logger.error(f"Failed on deploy: [HTTP {resp.status_code}] {resp.text}")
        else:
            raise Exception(
                f"Weird, this should never have happened... Your account is in state: {state}. {infostring}"
            )
    else:
        resp_dict = json.loads(resp.text)
        try:
            logger.error(
                f'Failed on get_status: [HTTP {resp.status_code}] {resp_dict["error"]} - {resp_dict["error_description"]}'
            )
        except Exception:
            logger.error(f"Failed on get_status: [HTTP {resp.status_code}] {resp.text}")
    raise Exception("Failed to get ssh username")


def is_valid_mc_url(mc_endpoint, verify=True):
    """make sure you always set the url schema when calling this method.
    This should be http or https
    """
    try:
        logger.info(f"Looking for motley_cue service at '{mc_endpoint}'...")
        parse_result = urlparse(mc_endpoint)
        fqdn_host = socket.getfqdn(parse_result.host)
        if fqdn_host and fqdn_host != parse_result.host:
            mc_endpoint = parse_result.copy_with(host=fqdn_host).unsplit()
            logger.info(f"Using FQDN for host: {mc_endpoint}")

        # a timeout is necessary here e.g. when the firewall drops packages
        resp = requests.get(mc_endpoint, verify=verify, timeout=TIMEOUT)

        from_cache = getattr(resp, "from_cache", False)
        if from_cache:
            logger.debug(f"Using cached response for {mc_endpoint}")

        if resp.status_code == 200:
            if not verify:
                logger.warning(
                    "InsecureRequestWarning: Unverified HTTPS"
                    f"request is being made to '{mc_endpoint}'. "
                    "Adding certificate verification is strongly advised."
                )
            # check for motley_cue
            if (
                resp.json().get("description", None)
                == "This is the user API for mapping remote identities to local identities."
            ):
                logger.info("...FOUND IT!")
                return mc_endpoint
    except requests.exceptions.SSLError:
        msg = (
            "SSL certificate verification failed. "
            "Use --insecure if you wish to ignore SSL certificate verification"
        )
        logger.info(msg)
        raise Exception(msg)
    except Exception as e:
        pass
        # logger.debug(f"Something went wrong: {e}")
    logger.info("...NOTHING HERE")
    return None



import struct
import fcntl
import termios
import signal
import sys
import re
import pexpect
import os
from functools import partial
from click import echo
from random import randint

from .logging import logger


PASSWORD_REGEX = r"(?:[^\n]*)(?:Access Token:)$"
SSH_HOSTNAME_PATTERN = re.compile(r"^hostname\s+(?P<hostname>\S+)\s+$", flags=re.MULTILINE)


def ssh_wrap(ssh_args, username, token, str_get_token=None, dry_run=False):
    """Runs the ssh command given by list of ssh_args, using given username
    and given token as password.

    When dry_run is true, it only prints the sshpass command; when the string
    representation of the command to get the token is not defined (str_get_token),
    the actual token is printed.
    """
    # add oidc-agent forwarding
    random_no = randint(10000, 99999)
    oidc_sock = os.getenv("OIDC_SOCK")
    if oidc_sock:
        ssh_args = ["-R", f"/tmp/oidc-forward-{random_no}:{oidc_sock}"] + ssh_args
    ssh_command_str = " ".join(ssh_args)
    ssh_command_str = f"ssh -l {username} {ssh_command_str}"
    if dry_run:
        __dry_run(ssh_command_str, tokens=token, str_get_tokens=str_get_token)
    else:
        __process_wrap(ssh_command_str, passwords=[token])


def scp_wrap(
    scp_args,
    username=None,
    tokens=None,
    str_get_tokens=None,
    num_prompts=1,
    dry_run=False,
):
    """Runs the scp command given by list of scp_args.

    If `username` and `tokens` are both None, we are in the NO_MOTLEY_CUE case
    and the process is simply started as it is; the user will interact with the process
    for any needed authentication; there is no motley_cue handling.

    If `username` is set, we are in the SINGLE_REMOTE case, where there is a
    single motley_cue instance in the arguments, so the user can be passed as an ssh
    option with '-o User'. The `tokens` argument must be a string in this case.
    The command will prompt for the token `num_prompts` times.

    If `username` is not set, but `tokens` is (must be a list), then we are in the MULTIPLE_REMOTES case,
    where the scp_args have been already augmented with usernames obtained from the motley_cue
    service(s). The `tokens` have to be input when prompted, in the given order. There should
    not be more prompts than tokens; any subsequent prompts will be forwarded to the user to handle.

    When dry_run is true, it only prints the needed command(s); if `str_get_tokens` is set,
    it prints the string representation(s) of the command(s) to get the token(s) instead
    of the actual token(s).
    """
    if not username and not tokens:
        passwords = tokens
    elif username and isinstance(tokens, str):
        scp_args = ["-o", f"User={username}"] + scp_args  # scp_args is a tuple
        passwords = [tokens] * num_prompts
    elif not username and isinstance(tokens, list):
        passwords = tokens
    else:
        raise Exception("Unsupported use case")

    scp_command_str = f"scp {' '.join(scp_args)}"
    if dry_run:
        __dry_run(
            scp_command_str,
            tokens=tokens,
            str_get_tokens=str_get_tokens,
            num_prompts=num_prompts,
        )
    else:
        __process_wrap(scp_command_str, passwords=passwords)


def get_hostname(ssh_args):
    """Try to get the ssh host from `ssh_args`
    by executing the ssh command with `-G` option
    and parsing the output for the actual HOSTNAME.
    """
    # add -G option to make ssh print its configuration
    # option added to the beginning of the list to avoid clashes with
    #   parameters from command to be executed remotely, e.g.:
    #   `ssh host ls -l`
    ssh_args = ssh_args.copy()
    if "-G" not in ssh_args:
        ssh_args.insert(0, "-G")

    command = f"ssh {' '.join(ssh_args)}"

    try:
        logger.debug(f"Running this command to get ssh configuration: {command}")
        output = pexpect.run(command).decode("utf-8")
        pattern_match = SSH_HOSTNAME_PATTERN.search(output)
        if not pattern_match:
            logger.error(f"Could not find hostname from ssh command {command}")
            return None
        hostname = pattern_match.group("hostname")
        logger.debug(f"Found hostname by parsing command output: {hostname}")
        return hostname
    except pexpect.ExceptionPexpect as e:
        logger.debug(e)
        logger.error(f"Error trying to get real hostname from ssh command {command}")
    return None


def __sigwinch_passthrough(sig=None, data=None, child_process=None):
    """Pass window changes to child"""
    s = struct.pack("HHHH", 0, 0, 0, 0)
    a = struct.unpack("hhhh", fcntl.ioctl(sys.stdout.fileno(), termios.TIOCGWINSZ, s))
    if child_process is not None and not child_process.closed:
        child_process.setwinsize(a[0], a[1])


def __output_filter(data, info=None):
    """Checks output from child process for Access Token prompt
    and sends the first password in list to the process.
    Removes the password from the list until the list is empty.
    """
    if (
        info
        and info["passwords"]
        and len(info["passwords"])
        and re.match(PASSWORD_REGEX, data.decode("utf-8"))
    ):
        info["child_process"].sendline(info["passwords"][0])
        info["child_process"].readline()  # to hide the token
        del info["passwords"][0]
        return b""
    return data


def __process_wrap(command, passwords=None):
    """Spawns a new process to run given command,
    and lets the user interact with it, except when prompted for
    Access Tokens, when it inputs the given passwords on
    behalf of the user, in the given order.
    """
    try:
        child_process = pexpect.spawn(command)
        signal.signal(
            signal.SIGWINCH,
            partial(__sigwinch_passthrough, child_process=child_process),
        )
        __sigwinch_passthrough(child_process=child_process)
        if passwords:
            info = {"child_process": child_process, "passwords": passwords}
            child_process.interact(output_filter=partial(__output_filter, info=info))
        else:
            child_process.interact()
    except pexpect.ExceptionPexpect as e:
        child_process.logout()
        logger.error(e)
    except Exception as e:
        logger.error(e)


def __dry_run(command, tokens=None, str_get_tokens=None, num_prompts=1):
    """Print string representation of the given ssh/scp `command`.

    If `tokens` is None, print `command` as it is.

    If `tokens` is a list, print the tokens first (or the string
    representation of the commands used to get the tokens, if
    `str_get_tokens` is set).

    If `tokens` is a string and only prompted for once, print the
    sshpass command to run the given ssh/scp command and to
    pass the access token when prompted for it. With multiple prompts,
    print the token and the command separately, with information.
    """
    if not str_get_tokens:
        str_get_tokens = tokens
    if not tokens:
        echo(command)
    elif isinstance(tokens, list):
        echo("# you'll need to input the tokens below when prompted, in this order:")
        for str_at in str_get_tokens:
            echo(f"echo {str_at}")
        echo(f"{command}")
    else:
        if num_prompts == 1:
            echo(f"SSHPASS={str_get_tokens} sshpass -P 'Access Token' -e {command}")
        else:
            echo(f"# you'll need to input the token below {num_prompts} times:")
            echo(f"echo {str_get_tokens}")
            echo(f"{command}")



import logging
import click_logging


logger = logging.getLogger(__name__)
echo_kwargs = {
    "debug": dict(err=True),
    "info": dict(err=True),
    "warning": dict(err=True),
    "error": dict(err=True),
    "exception": dict(err=True),
    "critical": dict(err=True),
}
style_kwargs = {
    "info": dict(fg="green"),
}
click_logging.basic_config(logger, echo_kwargs=echo_kwargs, style_kwargs=style_kwargs)



from rfc3986 import urlparse
from urllib import parse
import re
from click import UsageError
from enum import Enum

from .logging import logger


class ScpOperand:
    def __init__(
        self,
        remote=False,
        uri=False,
        user=None,
        host=None,
        path=None,
        port=None,
        original_str=None,
    ):
        self.__remote = remote
        self.__uri = uri
        self.__user = user
        self.__host = host
        self.__path = path
        self.__port = port
        self.__original_str = original_str

    @property
    def remote(self):
        """whether it is a remote path"""
        return self.__remote

    @property
    def uri(self):
        """whether the path is specified in URI form cf. RFCRFC3986"""
        return self.__uri

    @property
    def user(self):
        """remote user, if specified, or None if it is a local path"""
        return self.__user

    @property
    def host(self):
        """remote hostname or None if it is a local path"""
        return self.__host

    @property
    def path(self):
        """path, local or on remote host"""
        return self.__path

    @property
    def port(self):
        """port for SSH service"""
        return self.__port

    @property
    def original_str(self):
        """original string that was parsed into this class"""
        return self.__original_str

    def unsplit(self, user):
        if not self.remote:
            return self.path
        elif not self.uri:
            # format [user@]host[:path]
            if self.port:
                # should not happen
                raise Exception("Port cannot be specified in this format [user@]host[:path]")
            if not user:
                return f"{self.host}:{self.path}"
            else:
                return f"{user}@{self.host}:{self.path}"
        else:
            # format scp://[user@]host[:port][/path]
            return urlparse(self.original_str).copy_with(userinfo=user).unsplit()

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return (
            "ScpOperand["
            f"remote: {self.remote}, uri: {self.uri}, "
            f"user: {self.user}, host: {self.host}, "
            f"path: {self.path}, port: {self.port}"
            "]"
        )


class ScpUseCase(Enum):
    """Possible cases when running scp:

    * case 1: no motley_cue handling needed
        - all local paths, or
        - all remotes have user specified
    * case 2: a single motley_cue remote host
        - can also appear multiple times
        - the user can be passed as an ssh option with '-o'
        - no other remotes with specified users are present
    * case 3: multiple remotes
        - the sources/target have to be modified to pass the username
    """

    NO_MOTLEY_CUE = (1,)
    SINGLE_REMOTE = (2,)
    MULTIPLE_REMOTES = 3


class ScpCommand:
    def __init__(self, opts, sources, target):
        self.__opts = opts
        self.__sources = sources
        self.__target = target
        # default values
        self.__mc_host = None
        self.__num_prompts = 0
        # check which use case we're talking about
        number_of_remotes_with_user = 0
        number_of_unique_mc_remotes = 0
        number_of_pass_prompts = 0
        mc_host = None
        for operand in sources + [target]:
            if operand.remote and operand.user:
                # user is already set, do not handle via motley_cue
                number_of_remotes_with_user += 1
            elif operand.remote and operand.user is None:
                # this is definitely a motley_cue managed host
                if operand.host != mc_host:
                    # count unique mc remotes in command and reset pass prompts
                    number_of_unique_mc_remotes += 1
                    mc_host = operand.host
                    number_of_pass_prompts = 1
                else:
                    # increase number of password prompts for latest mc_host
                    number_of_pass_prompts += 1
        # set mc_host and pass_prompts only for SINGLE_REMOTE case
        if number_of_unique_mc_remotes == 0:
            self.__use_case = ScpUseCase.NO_MOTLEY_CUE
        elif number_of_unique_mc_remotes == 1 and not number_of_remotes_with_user:
            self.__use_case = ScpUseCase.SINGLE_REMOTE
            self.__mc_host = mc_host
            self.__num_prompts = number_of_pass_prompts
        else:
            self.__use_case = ScpUseCase.MULTIPLE_REMOTES

    @property
    def use_case(self):
        return self.__use_case

    @property
    def opts(self):
        return self.__opts

    @property
    def sources(self):
        return self.__sources

    @property
    def target(self):
        return self.__target

    @property
    def mc_host(self):
        return self.__mc_host

    @property
    def num_prompts(self):
        return self.__num_prompts

    def no_mc(self):
        return self.use_case == ScpUseCase.NO_MOTLEY_CUE

    def single_mc(self):
        return self.use_case == ScpUseCase.SINGLE_REMOTE

    def multiple_mc(self):
        return self.use_case == ScpUseCase.MULTIPLE_REMOTES

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return (
            "ScpCommand{\n"
            f"  use_case: {self.use_case}\n"
            f"  opts: {self.opts}\n"
            f"  sources: {self.sources}\n"
            f"  target: {self.target}\n"
            f"  mc_host: {self.mc_host}\n"
            f"  num_prompts: {self.num_prompts}\n"
            "}"
        )


def parse_scp_args(scp_args):
    """Extract source and target from the arguments
    of an scp command, represented as a list, and return
    an ScpCommand.
    """
    SCP_FLAGS = r"-[346ABCpqrTv]+"
    SCP_OPTS = r"-[cFiJloPS]"
    is_opt = False
    scp_opts = []
    operands = []
    for opt in scp_args:
        if opt.startswith("-") and re.match(SCP_FLAGS, opt):
            is_opt = False
            scp_opts += [opt]
        elif opt.startswith("-") and re.match(SCP_OPTS, opt):
            is_opt = True
            scp_opts += [opt]
        elif opt.startswith("-"):
            raise Exception(f"Invalid scp option: {opt}")
        elif is_opt:
            is_opt = False
            scp_opts += [opt]
        else:
            operands += [opt]

    if len(operands) < 2:
        raise UsageError("Please specify at least one SOURCE and TARGET for scp.")
    # target should be last
    target = __valid_path(operands[-1])
    # parse sources
    sources = [__valid_path(src) for src in operands[:-1]]
    return ScpCommand(opts=scp_opts, sources=sources, target=target)


def __valid_path(value):
    """Validates an scp operand (source/target) and returns ScpOperand
        - checks if it is a remote path
        - nothing to do for local paths
        - otherwise, parses the remote path and extracts info
    Parsing the remote path is similar to scp:
        - first try URI according to RFC3986
            scp://[user@]host[:port][/path]
        - otherwise remote path of the form:
            [user@]host:[path]
    """
    colon = __colon(value)
    if colon == -1:
        return ScpOperand(path=value, original_str=value)
    elif value.startswith("scp://"):
        logger.debug(f"{value} is of form scp://[user@]host[:port][/path]")
        try:
            parsed_uri = urlparse(value)
        except Exception as e:
            logger.debug(e)
            raise Exception(f"SCP operand {value} in URI form (cf RFC3986) could not be parsed")
        if not parsed_uri.host or parsed_uri.host == "":
            raise Exception(f"SCP operand {value} in URI form (cf RFC3986) does not contain a host")
        logger.debug(f"{parsed_uri}")
        return ScpOperand(
            remote=True,
            uri=True,
            user=parse.unquote(parsed_uri.userinfo)
            if parsed_uri.userinfo and parsed_uri.userinfo != ""
            else None,
            host=parse.unquote(parsed_uri.host),
            path=parse.unquote(parsed_uri.path)
            if parsed_uri.path and parsed_uri.path != ""
            else None,
            port=parsed_uri.port,
            original_str=value,
        )
    else:
        logger.debug(f"{value} is of form [user@]host:[path]")
        if colon == len(value):
            path = "."
        else:
            path = value[colon + 1 :]
        user_host = value[:colon].split("@")
        host = user_host[-1]
        user = "@".join(user_host[:-1])
        user = None if user == "" else user
        if host == "":
            raise Exception(f"SCP operand {value} in [user@]host:[path] does not contain a host")
        logger.debug(f"ParseResult: [user={user}, host={host}, path={path}]")
        return ScpOperand(remote=True, user=user, host=host, path=path, original_str=value)


def __colon(value):
    """This is the equivalent of the colon function in SCP

    Returns colon index or -1 when the host is not remote.
    """
    flag = False
    if value.startswith(":"):
        return -1
    if value.startswith("["):
        flag = True
    for i, v in enumerate(value):
        if v == "@" and len(value) > i + 1 and value[i + 1] == "[":
            flag = True
        if v == "]" and len(value) > i + 1 and value[i + 1] == ":" and flag:
            return i + 1
        if v == ":" and not flag:
            return i
        if v == "/":
            return -1
    return -1


# def extract_hostname_from_ssh(ssh_args):
#     """Extract the hostname from the arguments
#     of an ssh command, represented as a list
#     """
#     SSH_FLAGS = r"-[46AaCfGgKkMNnqsTtVvXxYy]+"
#     SSH_OPTS = r"-[BbcDEeFIiJLlmOopQRSWw]"
#     is_opt = False
#     for opt in ssh_args:
#         if opt.startswith("-") and re.match(SSH_FLAGS, opt):
#             is_opt = False
#         elif opt.startswith("-") and re.match(SSH_OPTS, opt):
#             is_opt = True
#         elif opt.startswith("-"):
#             raise Exception(f"Invalid ssh option: {opt}")
#         elif is_opt:
#             is_opt = False
#         else:
#             return opt



import typing as t
from gettext import gettext as _
import click
import click_logging
from click_option_group import optgroup, MutuallyExclusiveOptionGroup
from functools import wraps
import urllib3
import logging
import inspect

from .logging import logger
from .version import __version__


FC = t.TypeVar("FC", t.Callable[..., t.Any], click.Command)


class CustomUsageCommand(click.Command):
    @property
    def usage_text(self):
        return ""

    def format_usage(self, ctx, formatter):
        formatter.write(self.usage_text)
        formatter.write_paragraph()


class SshUsageCommand(CustomUsageCommand):
    @property
    def usage_text(self):
        return "mccli [OPTIONS] ssh [SSH OPTIONS] HOSTNAME [COMMAND]"


class ScpUsageCommand(CustomUsageCommand):
    @property
    def usage_text(self):
        return "mccli [OPTIONS] scp [SCP_OPTIONS] SOURCE ... TARGET"


def access_token_sources(func):
    @optgroup.group(
        "Access Token sources",
        help="The sources for retrieving an Access Token, in the order they are checked. If no source is specified, it will try to retrieve the supported token issuer from the service.",
        cls=MutuallyExclusiveOptionGroup,
    )
    @optgroup.option(
        "--token",
        metavar="TOKEN",
        envvar=[
            "ACCESS_TOKEN",
            "OIDC",
            "OS_ACCESS_TOKEN",
            "OIDC_ACCESS_TOKEN",
            "WATTS_TOKEN",
            "WATTSON_TOKEN",
        ],
        show_envvar=True,
        callback=validate_pass_from_parent,
        help="Pass token directly. Environment variables are checked in given order.",
    )
    @optgroup.option(
        "--oa-account",
        "--oidc",
        metavar="SHORTNAME",
        envvar=["OIDC_AGENT_ACCOUNT"],
        show_envvar=True,
        callback=validate_pass_from_parent,
        help="Name of configured account in oidc-agent.",
    )
    @optgroup.option(
        "--iss",
        "--issuer",
        metavar="URL",
        envvar=["OIDC_ISS", "OIDC_ISSUER"],
        show_envvar=True,
        callback=validate_pass_from_parent,
        help="URL of token issuer. Configured account in oidc-agent for this issuer will be used. Environment variables are checked in given order.",
    )
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def motley_cue_options(func):
    @optgroup.group("motley_cue options")
    @optgroup.option(
        "--mc-endpoint",
        metavar="URL",
        callback=validate_pass_from_parent,
        help="motley_cue API endpoint. Default URLs are checked in given order: https://HOSTNAME, https://HOSTNAME:8443, http://HOSTNAME:8080",
    )
    @optgroup.option(
        "--insecure",
        "verify",
        is_flag=True,
        default=True,
        callback=validate_verify,
        help="Ignore verifying the SSL certificate for motley_cue endpoint, NOT RECOMMENDED.",
    )
    @optgroup.option(
        "--no-cache",
        is_flag=True,
        default=False,
        callback=validate_pass_from_parent,
        help="Do not cache HTTP requests.",
    )
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def verbosity_options(func):
    @optgroup.group("Verbosity")
    @my_debug_option(logger)
    @my_logging_simple_verbosity_option(
        logger,
        "--log-level",
        default="ERROR",
        metavar="LEVEL",
        envvar="LOG",
        show_envvar=True,
    )
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def help_options(func):
    @optgroup.group("Help")
    @my_help_option("-h", "--help")
    @my_version_option(__version__, "-V", "--version")
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def basic_options(func):
    """The basic set of common options to be used by all commands.

    It still contains a hidden `--dry-run` option so that the user
    can pass it to the parent command as well.
    """

    @access_token_sources
    @motley_cue_options
    @verbosity_options
    @optgroup.option("--dry-run", is_flag=True, callback=validate_pass_from_parent, hidden=True)
    @help_options
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def extended_options(func):
    """Extends the basic common options by adding a `--dry-run` option to the verbosity group."""

    @access_token_sources
    @motley_cue_options
    @verbosity_options
    @optgroup.option(
        "--dry-run",
        is_flag=True,
        callback=validate_pass_from_parent,
        help="Print sshpass command and exit.",
    )
    @help_options
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def validate_verify(ctx, param, value):
    """
    First, take over value from parent, if set.
    Disable warnings when insecure is set.
    Might seem weird: the flag is called --insecure, but the
    meaning of the value is verify:
    - True by default, verify if HTTPS requests are secure, verify certificates
    - False means do not verify, disable warnings.
    When the flag is set, verify will be False.
    """
    try:
        value = ctx.meta[param.name]
    except Exception:
        # set the verify in the context meta dict to be used by subcommands
        # only if it was set through the commandline
        if ctx.get_parameter_source(param.name) is click.core.ParameterSource.COMMANDLINE:
            ctx.meta[param.name] = value
    if not value:
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
    return value


def validate_pass_from_parent(ctx, param, value):
    """If a param's value is set in the parent command,
    (via ctx.meta[param]) then take it over to the subcommand
    Otherwise, use given value and set it in the context
    to be propagated to subcommands
    """
    try:
        parent_value = ctx.meta[param.name]
        if parent_value:
            return parent_value
    except Exception:
        pass
    # set meta value to be used by subcommands only if it was set using the cmdline
    # envvar values should only be processed at subcommand level and not passed from parent
    # this way, the cmdline values take precedence over env vars or default values
    if value and ctx.get_parameter_source(param.name) is click.core.ParameterSource.COMMANDLINE:
        ctx.meta[param.name] = value
    return value


def my_logging_simple_verbosity_option(logger=None, *names, **kwargs):
    """My version of @click_logging.simple_verbosity_option
    that takes over the value from the parent command.

    A decorator that adds a `--verbosity, -v` option to the decorated
    command.

    Name can be configured through ``*names``. Keyword arguments are passed to
    the underlying ``click.option`` decorator.
    """
    if not names:
        names = ["--verbosity", "-v"]

    kwargs.setdefault("default", "INFO")
    kwargs.setdefault("metavar", "LVL")
    kwargs.setdefault("expose_value", False)
    kwargs.setdefault(
        "help",
        f'Either CRITICAL, ERROR, WARNING, INFO or DEBUG. Default value: {kwargs["default"]}.',
    )
    kwargs.setdefault("is_eager", True)

    logger = click_logging.core._normalize_logger(logger)

    def decorator(f):
        def _set_level(ctx, param, value):
            value = value.upper()
            # check if log_level was set in the parent command and use it as default
            try:
                value = ctx.meta["log_level"]
            except Exception as e:
                # set the log_level in the context meta dict to be used by subcommands
                # only if it was set through the commandline
                if ctx.get_parameter_source(param.name) is click.core.ParameterSource.COMMANDLINE:
                    ctx.meta["log_level"] = value

            x = getattr(logging, value, None)
            if x is None:
                raise click.BadParameter(
                    "Must be CRITICAL, ERROR, WARNING, INFO or DEBUG, not {}".format(value)
                )
            logger.setLevel(x)

        return optgroup.option(*names, callback=_set_level, **kwargs)(f)

    return decorator


def my_debug_option(logger=None, *names, **kwargs):
    """A decorator that adds a `--debug` option to the decorated command.

    Name can be configured through ``*names``. Keyword arguments are passed to
    the underlying ``click.option`` decorator.
    """
    if not names:
        names = ["--debug"]

    kwargs.setdefault("default", False)
    kwargs.setdefault("is_flag", True)
    kwargs.setdefault("expose_value", False)
    kwargs.setdefault("help", "Sets the log level to DEBUG.")
    kwargs.setdefault("is_eager", True)

    logger = click_logging.core._normalize_logger(logger)

    def decorator(f):
        def _set_debug(ctx, param, value):
            if value:
                # this option overwrites any parent option or --log-level options
                # when enabled, log level is always debug
                ctx.meta["log_level"] = "DEBUG"
                logger.setLevel(logging.DEBUG)

        return optgroup.option(*names, callback=_set_debug, **kwargs)(f)

    return decorator


def my_help_option(*param_decls: str, **kwargs: t.Any) -> t.Callable[[FC], FC]:
    """My version of @click.help_option that adds --help to the optgroup.

    Add a ``--help`` option which immediately prints the help page
    and exits the program.

    This is usually unnecessary, as the ``--help`` option is added to
    each command automatically unless ``add_help_option=False`` is
    passed.

    :param param_decls: One or more option names. Defaults to the single
        value ``"--help"``.
    :param kwargs: Extra arguments are passed to :func:`option`.
    """

    def callback(ctx: click.Context, param: click.Parameter, value: bool) -> None:
        if not value or ctx.resilient_parsing:
            return

        click.echo(ctx.get_help(), color=ctx.color)
        ctx.exit()

    if not param_decls:
        param_decls = ("--help",)

    kwargs.setdefault("is_flag", True)
    kwargs.setdefault("expose_value", False)
    kwargs.setdefault("is_eager", True)
    kwargs.setdefault("help", _("Show this message and exit."))
    kwargs["callback"] = callback
    return optgroup.option(*param_decls, **kwargs)


def my_version_option(
    version: t.Optional[str] = None,
    *param_decls: str,
    package_name: t.Optional[str] = None,
    prog_name: t.Optional[str] = None,
    message: t.Optional[str] = None,
    **kwargs: t.Any,
) -> t.Callable[[FC], FC]:
    """My version of @click.version_option that adds --version to
    the optgroup.

    Add a ``--version`` option which immediately prints the version
    number and exits the program.

    If ``version`` is not provided, Click will try to detect it using
    :func:`importlib.metadata.version` to get the version for the
    ``package_name``. On Python < 3.8, the ``importlib_metadata``
    backport must be installed.

    If ``package_name`` is not provided, Click will try to detect it by
    inspecting the stack frames. This will be used to detect the
    version, so it must match the name of the installed package.

    :param version: The version number to show. If not provided, Click
        will try to detect it.
    :param param_decls: One or more option names. Defaults to the single
        value ``"--version"``.
    :param package_name: The package name to detect the version from. If
        not provided, Click will try to detect it.
    :param prog_name: The name of the CLI to show in the message. If not
        provided, it will be detected from the command.
    :param message: The message to show. The values ``%(prog)s``,
        ``%(package)s``, and ``%(version)s`` are available. Defaults to
        ``"%(prog)s, version %(version)s"``.
    :param kwargs: Extra arguments are passed to :func:`option`.
    :raise RuntimeError: ``version`` could not be detected.
    """
    if message is None:
        message = _("%(prog)s, version %(version)s")

    if version is None and package_name is None:
        frame = inspect.currentframe()
        assert frame is not None
        assert frame.f_back is not None
        f_globals = frame.f_back.f_globals if frame is not None else None
        # break reference cycle
        # https://docs.python.org/3/library/inspect.html#the-interpreter-stack
        del frame

        if f_globals is not None:
            package_name = f_globals.get("__name__")

            if package_name == "__main__":
                package_name = f_globals.get("__package__")

            if package_name:
                package_name = package_name.partition(".")[0]

    def callback(ctx: click.Context, param: click.Parameter, value: bool) -> None:
        if not value or ctx.resilient_parsing:
            return

        nonlocal prog_name
        nonlocal version

        if prog_name is None:
            prog_name = ctx.find_root().info_name

        if version is None and package_name is not None:
            metadata: t.Optional[types.ModuleType]

            try:
                from importlib import metadata  # type: ignore
            except ImportError:
                # Python < 3.8
                import importlib_metadata as metadata  # type: ignore

            try:
                version = metadata.version(package_name)  # type: ignore
            except metadata.PackageNotFoundError:  # type: ignore
                raise RuntimeError(
                    f"{package_name!r} is not installed. Try passing" " 'package_name' instead."
                )

        if version is None:
            raise RuntimeError(
                f"Could not determine the version for {package_name!r} automatically."
            )

        click.echo(
            t.cast(str, message) % {"prog": prog_name, "package": package_name, "version": version},
            color=ctx.color,
        )
        ctx.exit()

    if not param_decls:
        param_decls = ("--version",)

    kwargs.setdefault("is_flag", True)
    kwargs.setdefault("expose_value", False)
    kwargs.setdefault("is_eager", True)
    kwargs.setdefault("help", _("Show the version and exit."))
    kwargs["callback"] = callback
    return optgroup.option(*param_decls, **kwargs)


def tuple_to_list(ctx, param, value):
    try:
        if not value:
            logger.error("Empty command args list. Shouldn't have happened...")
            ctx.abort()
        elif not isinstance(value, tuple):
            logger.error("Weird, command args is not a tuple...")
        else:
            return list(value)
    except Exception as e:
        logger.debug(e)
        logger.error("Failed to convert command args tuple to list")
        ctx.abort()



from flaat import BaseFlaat
from flaat.access_tokens import get_access_token_info
import json

from .motley_cue_client import get_info, get_authorisation_info, get_local_status


flaat = BaseFlaat()

flaat.set_verbosity(0)
# flaat.set_cache_lifetime(120) # seconds; default is 300

flaat.set_trusted_OP_list(
    [
        "https://b2access.eudat.eu/oauth2/",
        "https://b2access-integration.fz-juelich.de/oauth2",
        "https://unity.helmholtz-data-federation.de/oauth2/",
        "https://login.helmholtz-data-federation.de/oauth2/",
        "https://login-dev.helmholtz.de/oauth2/",
        "https://login.helmholtz.de/oauth2/",
        "https://unity.eudat-aai.fz-juelich.de/oauth2/",
        "https://services.humanbrainproject.eu/oidc/",
        "https://accounts.google.com/",
        "https://aai.egi.eu/oidc/",
        "https://aai.egi.eu/auth/realms/egi",
        "https://aai-demo.egi.eu/auth/realms/egi",
        "https://aai-demo.egi.eu/oidc/",
        "https://aai-dev.egi.eu/oidc/",
        "https://aai-dev.egi.eu/auth/realms/egi",
        "https://login.elixir-czech.org/oidc/",
        "https://iam-test.indigo-datacloud.eu/",
        "https://iam.deep-hybrid-datacloud.eu/",
        "https://iam.extreme-datacloud.eu/",
        "https://oidc.scc.kit.edu/auth/realms/kit/",
        "https://proxy.demo.eduteams.org",
        "https://wlcg.cloud.cnaf.infn.it/",
    ]
)


def get_all_info(mc_url, token, verify=False):
    if not mc_url and not token:
        return None

    info_string = ""
    if mc_url:
        service_info = get_info(mc_url, verify)
        if service_info:
            info_string += "\n==== Information about service ====\n"
            info_string += json.dumps(service_info, indent=4)
            info_string += "\n"
    if mc_url and token:
        authz_info = get_authorisation_info(mc_url, token, verify)
        if authz_info:
            info_string += "\n==== Authorisation on service for provided token issuer (OP) ====\n"
            info_string += json.dumps(authz_info, indent=4)
            info_string += "\n"
        local_status = get_local_status(mc_url, token, verify)
        if local_status:
            info_string += (
                "\n==== Information about your local account on service for provided token ====\n"
            )
            info_string += local_status
            info_string += "\n"
    if token:
        access_token_info = get_access_token_info(token)
        info_string += "\n==== Information stored inside the provided Access Token ====\n"
        if access_token_info:
            info_string += json.dumps(
                {
                    "header": access_token_info.header,
                    "body": access_token_info.body,
                    "signature": access_token_info.signature,
                },
                indent=4,
            )
        else:
            info_string += "Your access token is not a JWT. I.e. it does not contain information (at least I cannot find it.)"
        info_string += "\n"

        user_infos = flaat.get_user_infos_from_access_token(token)
        info_string += "\n==== Information retrieved from userinfo endpoint ====\n"
        if user_infos:
            info_string += json.dumps(user_infos.user_info, indent=4)
        else:
            info_string += "The response from the userinfo endpoint does not contain information (at least I cannot find it.)"
        info_string += "\n"

        if user_infos:
            timeleft = user_infos.valid_for_secs
            if timeleft is not None:
                if timeleft > 0:
                    info_string += "\nToken valid for %.1f more seconds." % timeleft
                else:
                    info_string += "\nYour token is already EXPIRED for %.1f seconds!" % abs(
                        timeleft
                    )
                info_string += "\n"

    return info_string



#!/usr/bin/env python3

import click

from .ssh_wrapper import ssh_wrap, scp_wrap
from .init_utils import (
    valid_mc_url,
    init_endpoint,
    init_token,
    init_user,
    init_cache,
    augmented_scp_command,
    check_and_replace_long_token,
)
from .scp_utils import parse_scp_args
from .click_utils import (
    SshUsageCommand,
    ScpUsageCommand,
    tuple_to_list,
    basic_options,
    extended_options,
)
from .info_utils import get_all_info
from .logging import logger


@click.group(invoke_without_command=False, add_help_option=False)
@basic_options
def cli(**kwargs):
    """
    SSH client wrapper with OIDC-based authentication
    """
    pass


@cli.command(name="info", add_help_option=False, short_help="get info about service")
@basic_options
@click.argument("hostname", required=False)
def info(mc_endpoint, verify, no_cache, token, oa_account, iss, dry_run, hostname):
    """Shows various information about the user and the service:

    If an Access Token is provided, show information in the token,
    as well as info about the user retrieved from the token issuer.

    If HOSTNAME is provided, show information about SSH service
    running on HOSTNAME: supported OIDC providers, service
    description and login info.

    If both token and HOSTNAME are provided, also show authorisation
    information if issuer of token is supported on the service, as well
    as the status of the local account on service.
    """
    try:
        if not no_cache:
            init_cache()
        if hostname is None:
            raise Exception("No HOSTNAME provided.")
        if mc_endpoint:
            mc_url = valid_mc_url(mc_endpoint, verify)
        else:
            mc_url = init_endpoint([hostname], verify)
    except Exception as e:
        mc_url = None
        logger.warning(e)
        logger.warning("Cannot show service-related information.")
    try:
        at, _ = init_token(
            token,
            oa_account,
            iss,
            mc_endpoint=mc_url,
            verify=verify,
            validate_length=False,
        )
    except Exception as e:
        at = None
        logger.warning(e)
        logger.warning("Cannot show token information")

    info_string = get_all_info(mc_url, at, verify)
    if info_string:
        click.echo(info_string)
    else:
        logger.error(
            "No information available: please provide a hostname and/or an Access Token.\n"
            + "Try 'mccli info --help' for usage information."
        )


@cli.command(
    name="ssh",
    short_help="remote login client",
    cls=SshUsageCommand,
    context_settings={"ignore_unknown_options": True, "allow_extra_args": True},
)
@extended_options
@click.argument(
    "ssh_command",
    nargs=-1,
    required=True,
    type=click.UNPROCESSED,
    callback=tuple_to_list,
)
def ssh(mc_endpoint, verify, no_cache, token, oa_account, iss, dry_run, ssh_command):
    """Connects and logs into HOSTNAME via SSH by using the provided OIDC
    Access Token to authenticate.

    If a COMMAND is specified, it is executed on the remote host instead
    of a login shell.

    The remote user must not be specified, since it will be obtained from
    the motley_cue service. Any specified username will be ignored.

    When no Access Token source is specified, the service on the remote host
    is queried for supported issuers; if only one issuer is supported,
    this is used to retrieve the token from the oidc-agent.
    """
    try:
        if not no_cache:
            init_cache()
        if mc_endpoint:
            mc_url = valid_mc_url(mc_endpoint, verify)
        else:
            mc_url = init_endpoint(ssh_command, verify)
        at, str_get_at = init_token(token, oa_account, iss, mc_endpoint=mc_url, verify=verify)
        username = init_user(mc_url, at, verify)
        at, str_get_at = check_and_replace_long_token(at, str_get_at)
        ssh_wrap(ssh_command, username, at, str_get_token=str_get_at, dry_run=dry_run)
    except Exception as e:
        logger.error(e)


@cli.command(
    name="scp",
    short_help="secure file copy",
    cls=ScpUsageCommand,
    context_settings={"ignore_unknown_options": True, "allow_extra_args": True},
)
@extended_options
@click.argument(
    "scp_command",
    nargs=-1,
    required=True,
    type=click.UNPROCESSED,
    callback=tuple_to_list,
)
def scp(mc_endpoint, verify, no_cache, token, oa_account, iss, dry_run, scp_command):
    """Copies files between hosts on a network over SSH using the provided
    OIDC Access Token to authenticate.

    The SOURCE and TARGET may be specified as a local pathname, a remote host
    with optional path in the form [user@]host:[path], or a URI in the form
    scp://[user@]host[:port][/path]

    The remote user should not be specified, since it will be obtained from
    the motley_cue service. If you specify a username for a host, then it
    will be used; it will be assumed that this specific host does not use
    motley_cue, and by extension, token authentication; you will have to handle
    authentication for this host on your own.

    When no Access Token source is specified, the remote host is queried for
    supported issuers; if only one issuer is supported, this is used to
    retrieve the token from the oidc-agent.
    """
    try:
        if not no_cache:
            init_cache()
        if mc_endpoint:
            mc_url = valid_mc_url(mc_endpoint, verify)
            at, str_get_at = init_token(token, oa_account, iss, mc_endpoint=mc_url, verify=verify)
            username = init_user(mc_url, at, verify)
            at, str_get_at = check_and_replace_long_token(at, str_get_at)
            scp_wrap(
                scp_command,
                username=username,
                tokens=at,
                str_get_tokens=str_get_at,
                dry_run=dry_run,
            )
        else:
            scp_args = parse_scp_args(scp_command)
            if scp_args.no_mc():
                logger.warning(
                    "No motley_cue handling will be done. "
                    "Either all specified paths are local, "
                    "or users are specified for remotes."
                )
                scp_wrap(scp_command, dry_run=dry_run)
            elif scp_args.single_mc():
                logger.info("Only one host with motley_cue detected. Easy.")
                mc_url = init_endpoint([scp_args.mc_host], verify)
                at, str_get_at = init_token(
                    token, oa_account, iss, mc_endpoint=mc_url, verify=verify
                )
                username = init_user(mc_url, at, verify)
                at, str_get_at = check_and_replace_long_token(at, str_get_at)
                scp_wrap(
                    scp_command,
                    username=username,
                    tokens=at,
                    str_get_tokens=str_get_at,
                    num_prompts=scp_args.num_prompts,
                    dry_run=dry_run,
                )
            elif scp_args.multiple_mc():
                logger.info(
                    "Multiple hosts with motley_cue detected, "
                    "your commandline will be augmented with usernames. "
                )
                new_scp_command, tokens, str_get_tokens = augmented_scp_command(
                    scp_args, token, oa_account, iss, verify
                )
                scp_wrap(
                    new_scp_command,
                    tokens=tokens,
                    str_get_tokens=str_get_tokens,
                    dry_run=dry_run,
                )
            else:
                raise Exception(
                    "Something went wrong when trying to find out "
                    "which paths are remote and which are local."
                )
    except Exception as e:
        logger.error(e)


@cli.command(name="sftp", short_help="secure file transfer")
def sftp():
    """
    --- Not implemented ---
    """
    logger.error("Not implemented.")


if __name__ == "__main__":
    cli()



from logging import log
import requests_cache
import hashlib
import os.path
import liboidcagent as agent
from time import time
from flaat.access_tokens import get_access_token_info
import requests
import json

from .motley_cue_client import (
    local_username,
    get_supported_ops,
    is_valid_mc_url,
    get_audience,
    generate_otp,
)
from .ssh_wrapper import get_hostname
from .logging import logger

# some predefined oidc-gen commands for different issuers
oidc_gen_command_strings = {
    "aai.egi.eu/oidc": 'oidc-gen --pub --iss https://aai.egi.eu/oidc --scope "openid profile email offline_access eduperson_entitlement eduperson_scoped_affiliation eduperson_unique_id" egi',
    "wlcg.cloud.cnaf.infn.it": 'oidc-gen --pub --issuer https://wlcg.cloud.cnaf.infn.it --scope "openid profile offline_access eduperson_entitlement eduperson_scoped_affiliation wlcg.groups wlcg" wlcg',
    "login.helmholtz.de/oauth2": 'oidc-gen --pub --iss https://login.helmholtz.de/oauth2 --scope "openid profile email offline_access eduperson_entitlement eduperson_scoped_affiliation eduperson_unique_id" helmholtz',
    "accounts.google.com": "oidc-gen --pub --iss https://accounts.google.com/ --flow device --scope max google",
}


def canonical_url(url):
    """Strip URL of protocol info and ending slashes"""
    url = url.lower()
    if url.startswith("http://"):
        url = url[7:]
    if url.startswith("https://"):
        url = url[8:]
    if url.startswith("www."):
        url = url[4:]
    if url.endswith("/"):
        url = url[:-1]
    return url


def oidc_gen_command(iss):
    """Return a string containing the appropriate oidc-gen command
    for given issuer, including suggested scopes.
    """
    return oidc_gen_command_strings.get(canonical_url(iss), f"oidc-gen --iss {iss}")


def check_and_replace_long_token(token, str_init_token):
    """If token too long, create a new OTP to be used as ssh password, by hashing given token."""
    if len(token) < 1024:
        return token, str_init_token
    otp = hashlib.sha512(bytearray(token, "ascii")).hexdigest()
    logger.debug(f"Created OTP [{otp}] to be used as SSH password.")
    return otp, otp


def _validate_token_length(func):
    """Decorator for init_token that checks if token length is < 1024

    The function takes token returned by init_token and raises an
    Exception if token too long and cannot be used for SSH authentication.
    """

    def wrapper(*args, **kwargs):
        at, str_get_at = func(*args, **kwargs)
        if kwargs.get("validate_length", True) and len(at) >= 1024:
            mc_endpoint = kwargs.get("mc_endpoint", None)
            verify = kwargs.get("verify", True)
            response = generate_otp(mc_endpoint=mc_endpoint, token=at, verify=verify)
            use_otp = False
            if response.status_code == requests.codes.ok:  # pylint: disable=no-member
                supported = response.json().get("supported", False)
                successful = response.json().get("successful", False)
                use_otp = supported and successful
                if not use_otp:
                    raise Exception(
                        f"Sorry, your token is too long ({len(at)} >= 1024) and cannot be used for SSH "
                        "authentication. Please ask your OP admin if they can release shorter tokens, "
                        "or the service admin if they can support one-time passwords."
                    )
                else:
                    logger.debug(
                        f"Generated one-time password for use with SSH instead of long access token."
                    )
            else:  # probably not authorised (401, 403), or internal server error (500)
                resp_dict = json.loads(response.text)
                try:
                    logger.error(
                        f'Failed on generate_otp: [HTTP {response.status_code}] {resp_dict["error"]} - {resp_dict["error_description"]}'
                    )
                except Exception:
                    logger.error(f"Failed on generate_otp: [HTTP {response.status_code}] {response.text}")
        return at, str_get_at

    return wrapper


def _get_access_token(oa_account=None, iss=None, mc_endpoint=None, verify=True):
    """Retrieve access token from oidc-agent, then query motley_cue API with this token
    to check if specific audience is needed for authz
    """

    def _get_token_from_agent(oa_account=None, iss=None, audience=None):
        if oa_account is not None:
            return agent.get_access_token(
                oa_account, application_hint="mccli", audience=audience
            ), _str_init_token(oa_account=oa_account, audience=audience)
        elif iss is not None:
            return agent.get_access_token_by_issuer_url(
                iss, application_hint="mccli", audience=audience
            ), _str_init_token(iss=iss, audience=audience)
        return None, None

    at, str_at = _get_token_from_agent(oa_account, iss)
    if at is not None:
        audience = get_audience(mc_endpoint, at, verify)
        if audience is not None:
            at, str_at = _get_token_from_agent(oa_account, iss, audience)
    return at, str_at


@_validate_token_length
def init_token(token, oa_account, iss, mc_endpoint=None, verify=True, validate_length=True):
    """Retrieve an oidc token:

    * use token if set,
    * retrieve from the oidc-agent via given account if oa_account is set
    * retrieve from the oidc-agent via given issuer if iss is set
    * use iss from service, if only one iss is supported
    * fail

    return token and string representation of command to retrieve token
    """
    expired = False
    if token is not None:
        # check if token is expired
        try:
            info_in_token = get_access_token_info(token)
            if info_in_token:
                timeleft = info_in_token.body["exp"] - time()
            else:
                timeleft = None
        except Exception:
            timeleft = None
        if not timeleft:
            logger.warning(
                "Could not get expiration date from provided token, it might not be a JWT. Using it anyway..."
            )
            logger.debug(f"Access Token: {token}")
            return token, _str_init_token(token=token)
        elif timeleft > 0:
            logger.info(f"Token valid for {timeleft} more seconds, using provided token.")
            logger.debug(f"Access Token: {token}")
            return token, _str_init_token(token=token)
        else:
            expired = True
            logger.warning(
                f"Token is expired for {0-timeleft} seconds. Looking for another source for Access Token..."
            )
            logger.debug(f"Access Token: {token}")
    else:
        logger.info("No access token provided.")
    if oa_account is not None:
        try:
            logger.info(f"Using oidc-agent account: {oa_account}")
            return _get_access_token(oa_account=oa_account, mc_endpoint=mc_endpoint)
        except Exception as e:
            logger.warning(
                f"Failed to get Access Token for oidc-agent account '{oa_account}': {e}."
            )
            logger.warning(
                f"Are you sure this account is loaded? Load it with:\n    oidc-add {oa_account}"
            )
            logger.warning(
                f"Are you sure this account is configured? Create it with:\n    oidc-gen {oa_account}"
            )
    else:
        logger.info("No oidc-agent account provided.")
    if iss is not None:
        try:
            logger.info(f"Using issuer: {iss}")
            if not iss.startswith("http"):
                iss = f"https://{iss}"
                logger.warning(
                    f"The issuer URL you provided does not contain protocol information, assuming HTTPS: {iss}"
                )
            return _get_access_token(iss=iss, mc_endpoint=mc_endpoint)
        except Exception as e:
            logger.warning(f"Failed to get Access Token from oidc-agent for issuer '{iss}': {e}.")
            logger.warning(
                f"Are you sure the issuer URL is correct or that you have an account configured with oidc-agent for this issuer? Create it with:\n    {oidc_gen_command(iss)}"
            )
    else:
        logger.info("No issuer URL provided.")
    if mc_endpoint is not None:
        logger.info(f"Trying to get list of supported AT issuers from {mc_endpoint}...")
        supported_ops = get_supported_ops(mc_endpoint, verify)
        if len(supported_ops) == 1:
            iss = supported_ops[0]
            try:
                logger.info(
                    f"Using the only issuer supported on service to retrieve token from oidc-agent: {iss}"
                )
                return _get_access_token(iss=iss, mc_endpoint=mc_endpoint)
            except Exception as e:
                logger.warning(
                    f"Failed to get Access Token from oidc-agent for the only issuer supported on service '{iss}': {e}"
                )
                logger.warning(
                    f"If you don't have an oidc-agent account configured for this issuer, create it with:\n    {oidc_gen_command(iss)}"
                )
        elif len(supported_ops) > 1:
            logger.warning("Multiple issuers supported on service, I don't know which one to use:")
            logger.warning("[" + "\n    ".join([""] + supported_ops) + "\n]")
    if expired:
        msg = (
            "The provided Access Token is expired. Have you considered using 'oidc-agent' to always have valid tokens?\n"
            + "    https://github.com/indigo-dc/oidc-agent\n"
            + "Try 'mccli --help' for help on specifying the Access Token source."
        )
    else:
        msg = (
            "No Access Token found.\n"
            + "Try 'mccli --help' for help on specifying the Access Token source."
        )
    raise Exception(msg)


def _str_init_token(token=None, oa_account=None, iss=None, audience=None):
    """String representation of command used to get Access Token:
    * full token if token is set
    * `oidc-token oa_account` if oidc-agent account is set
    * `oidc-token iss` if issuer is set
    *  ... (`oidc-token iss` if iss can be retrieved from service)
    """
    aud_str = ""
    if audience:
        aud_str = f"--aud {audience} "
    if token:
        return f"'{token}'"
    if oa_account:
        return f"`oidc-token {aud_str}{oa_account}`"
    elif iss:
        return f"`oidc-token {aud_str}{iss}`"
    raise Exception("No access token found")


def valid_mc_url(mc_endpoint, verify=True):
    """Checks if there is a motley_cue service
    running at provided url and returns the url
    Raises an exception otherwise.
    """
    if mc_endpoint.startswith("http"):
        valid_endpoint = is_valid_mc_url(mc_endpoint, verify)
        if valid_endpoint:
            return valid_endpoint
    else:
        for schema in ["http", "https"]:
            # they should be in this order, since https can raise SSL related exception
            logger.warning(f"No URL schema specified for mc-endpoint, trying {schema}")
            endpoint = f"{schema}://{mc_endpoint}"
            valid_endpoint = is_valid_mc_url(endpoint, verify)
            if valid_endpoint:
                return valid_endpoint
    msg = (
        f"No motley_cue service found at '{mc_endpoint}'. "
        "Please specify a valid motley_cue endpoint."
    )
    raise Exception(msg)


def init_endpoint(ssh_args, verify=True):
    """Initialise motley_cue endpoint from ssh args.

    (HACKY) Try to get the ssh host from `ssh_args`
    by executing the ssh command with invalid `-b` option
    and parsing the output for the actual HOSTNAME.

    Then try to use default value for motley_cue endpoint: https://HOSTNAME
    If this is not reachable, issue warning and try: http://HOSTNAME:8080
    If also not reachable, exit and ask user to specify it using --mc-endpoint
    """
    logger.info("Trying to get ssh hostname from arguments.")
    ssh_host = get_hostname(ssh_args)
    if not ssh_host:  # raise error and ask user to specify endpoint
        msg = f"Could not resolve hostname."
        raise Exception(msg)
    logger.info(f"Got host '{ssh_host}', looking for motley_cue service on host.")

    # try https
    endpoint = f"https://{ssh_host}"
    valid_endpoint = is_valid_mc_url(endpoint, verify)
    if valid_endpoint:
        return valid_endpoint

    # try https on 8443
    endpoint = f"https://{ssh_host}:8443"
    valid_endpoint = is_valid_mc_url(endpoint, verify)
    if valid_endpoint:
        return valid_endpoint

    # try http on 8080 but issue warning
    endpoint = f"http://{ssh_host}:8080"
    valid_endpoint = is_valid_mc_url(endpoint)
    if valid_endpoint:
        logger.warning(f"using unencrypted motley_cue endpoint: {endpoint}")
        return valid_endpoint

    # raise error and ask user to specify endpoint
    msg = (
        f"No motley_cue service found on host '{ssh_host}' "
        "on port 443, 8443 or 8080. "
        "Please specify motley_cue endpoint via --mc-endpoint."
    )
    raise Exception(msg)


def init_user(mc_endpoint, token, verify=True):
    """Get remote username, will be deployed if it doesn't exist."""
    return local_username(mc_endpoint, token, verify=verify)


def augmented_scp_command(scp_command, token, oa_account, iss, verify=False):
    """Receives an ScpCommand object and adds the username for each
    host that is remote and has no user specified, by contacting
    the motley_cue service on that host.
    Returns new scp command as list, as well as the tokens that should be
    passed as passwords, in order. A third list contains the
    string representation of each command used to get each token.
    """
    scp_args = scp_command.opts
    tokens = []
    str_get_tokens = []
    for operand in scp_command.sources + [scp_command.target]:
        if operand.remote and operand.user is None:
            # this is definitely a motley_cue managed host
            logger.debug(f"Trying to get username from motley_cue service on {operand.host}.")
            mc_url = init_endpoint([operand.host], verify)
            logger.debug("mc endpoint: %s", mc_url)
            at, str_get_at = init_token(token, oa_account, iss, mc_endpoint=mc_url, verify=verify)
            username = init_user(mc_url, at, verify)
            at, str_get_at = check_and_replace_long_token(at, str_get_at)
            scp_args += [operand.unsplit(username)]
            tokens += [at]
            str_get_tokens += [str_get_at]
        else:
            # add the operand as it was given
            scp_args += [operand.original_str]

    return scp_args, tokens, str_get_tokens


def init_cache():
    """Function to install caching for HTTP requests.

    Make sure to call this function after setting log level
    for full output.
    """
    cache_name = os.path.expanduser("~/.cache/mccli_cache")
    expire_after = 300  # 5 min
    include_get_headers = True
    allowable_methods = "GET"
    urls_expire_after = {
        "*/user/*": 0,
        "*/admin/*": 0,
    }

    try:
        requests_cache.install_cache(
            cache_name=cache_name,
            expire_after=expire_after,
            allowable_methods=allowable_methods,
            include_get_headers=include_get_headers,
            urls_expire_after=urls_expire_after,
        )

        if requests_cache.is_installed():
            # requests_cache.remove_expired_responses()
            logger.debug(f"HTTP requests cache installed at {cache_name}.sqlite")
    except Exception as e:
        logger.debug(f"Could not install requests cache: {e}. Uninstalling cache...")
        requests_cache.uninstall_cache()
        logger.warning(
            f"Something went wrong when initialising cache, will not cache HTTP requests. Executing command might be slower."
        )
